import os
import pytest
import serial

from matterlab_serial_device import SerialDevice, open_close
from matterlab_serial_device.serial_device import (
    SERIAL_PARITY,
    SERIAL_BYTESIZE,
    SERIAL_STOPBITS,
)

# ATTN: Better tests could be written with the hypothesis library. Leaving for later.

PORT = "/dev/ttyUSB0" if os.name == "posix" else "COM1"


@pytest.fixture
def mock_serial(mocker):
    """


    Args:
      mocker:

    Returns:

    """
    mock_serial = mocker.patch("matterlab_serial_device.serial_device.serial.Serial", autospec=True)
    serial_instances = []
    mock_serial_instance = mocker.MagicMock()
    mock_serial_instance.com_port = PORT  # Set the expected return value
    mock_serial_instance._factory = mock_serial
    mock_serial_instance._instances = serial_instances
    mock_serial.return_value = mock_serial_instance

    def _make_serial_instance(*args, **kwargs):
        if not serial_instances:
            instance = mock_serial_instance
        else:
            instance = mocker.MagicMock()
            instance.com_port = kwargs.get("port", PORT)
        serial_instances.append(instance)
        return instance

    mock_serial.side_effect = _make_serial_instance
    yield mock_serial_instance


@pytest.fixture(autouse=True)
def clear_shared_devices():
    SerialDevice._shared_devices.clear()
    yield
    SerialDevice._shared_devices.clear()


@pytest.fixture
def device_fixture(mock_serial):
    """


    Args:
      mock_serial:

    Returns:

    """
    serial_device = SerialDevice(com_port=PORT)
    yield serial_device


def test_init_defaults(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    # Test class attributes
    assert device_fixture.default_eol == b"\r\n"

    # Test default instance attributes
    assert device_fixture.com_port == PORT
    assert device_fixture.encoding == "utf-8"
    assert device_fixture.baudrate == 9600
    assert device_fixture.parity == serial.PARITY_NONE
    assert device_fixture.bytesize == serial.EIGHTBITS
    assert device_fixture.timeout is None
    assert device_fixture.stopbits == serial.STOPBITS_ONE

    # Test that the mocked Serial class was called with the correct arguments
    assert device_fixture.device.com_port == PORT

    # Test that the COM port was closed after device initialization in SerialDevice._init_device()
    mock_serial.close.assert_called_once()


def test_init_bytesize(mock_serial):
    """


    Args:
      mock_serial:

    Returns:

    """
    for key, value in SERIAL_BYTESIZE.items():
        serial_device = SerialDevice(com_port=PORT, bytesize=key)
        assert serial_device.bytesize == value


def test_init_parity(mock_serial):
    """


    Args:
      mock_serial:

    Returns:

    """
    for key, value in SERIAL_PARITY.items():
        serial_device = SerialDevice(com_port=PORT, parity=key)
        assert serial_device.parity == value


def test_init_stopbits(mock_serial):
    """


    Args:
      mock_serial:

    Returns:

    """
    for key, value in SERIAL_STOPBITS.items():
        serial_device = SerialDevice(com_port=PORT, stopbits=key)
        assert serial_device.stopbits == value


def test_open_device_comm(mocker, mock_serial, device_fixture):
    """


    Args:
      mocker:
      mock_serial:
      device_fixture:

    Returns:

    """
    device_fixture.open_device_comm()

    # Define the expected call sequence
    expected_calls = [mocker.call.close(), mocker.call.close(), mocker.call.open()]

    # Check that the calls were made in the specified order
    mock_serial.assert_has_calls(expected_calls, any_order=False)


def test_close_device_comm(mocker, mock_serial, device_fixture):
    """


    Args:
      mocker:
      mock_serial:
      device_fixture:

    Returns:

    """
    device_fixture.close_device_comm()

    # Define the expected call sequence
    expected_calls = [mocker.call.close(), mocker.call.close()]

    # Check that the calls were made in the specified order
    mock_serial.assert_has_calls(expected_calls, any_order=False)


def test_open_close(mocker, mock_serial, device_fixture):
    """


    Args:
      mocker:
      mock_serial:
      device_fixture:

    Returns:

    """

    class DummyDevice(SerialDevice):
        """ """

        @open_close
        def _dummy_method(self):
            """ """
            pass

    dummy_device = DummyDevice(com_port=PORT)
    dummy_device._dummy_method()

    # Define the expected call sequence
    expected_calls = [
        mocker.call.close(),
        mocker.call.close(),
        mocker.call.open(),
        mocker.call.close(),
    ]

    # Check that the calls were made in the specified order
    dummy_device.device.assert_has_calls(expected_calls, any_order=False)


def test_encode(device_fixture):
    """


    Args:
      device_fixture:

    Returns:

    """
    assert device_fixture._encode("test") == b"test"
    assert device_fixture._encode(b"test") == b"test"
    assert device_fixture._encode(bytearray(b"test")) == b"test"
    with pytest.raises(TypeError):
        device_fixture._encode(123)


def test_write(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    device_fixture.write("test")
    device_fixture.device.write.assert_called_once_with(b"test")


def test_read(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    for return_bytes in [True, False]:
        mock_serial.read_until.return_value = b"test"
        response = device_fixture.read(return_bytes=return_bytes)

        if return_bytes:
            assert response == b"test"
        else:
            assert response == "test"

        device_fixture.device.read_until.assert_called_once_with(expected=b"\r\n", size=None)
        mock_serial.reset_mock()  # Reset call counter at end of loop


def test_read_bytes_default(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    expected_response = b"test"
    mock_serial.read_until.return_value = expected_response

    response = device_fixture.read_bytes()

    assert response == expected_response
    device_fixture.device.read_until.assert_called_once_with(expected=b"\r\n", size=None)


def test_read_bytes_until(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    for read_until in [b"end", "end"]:
        device_fixture.read_bytes(read_until=read_until)

        mock_serial.read_until.assert_called_once_with(expected=b"end", size=None)
        mock_serial.reset_mock()  # Reset call counter at end of loop


def test_read_bytes_num_bytes(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    device_fixture.read_bytes(num_bytes=10)

    mock_serial.read_until.assert_called_once_with(expected=b"\r\n", size=10)


def test_read_bytes_remove_from_start(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    mock_serial.read_until.return_value = b"test\r\n"
    expected_response = b"test\r\n"[2:]

    response = device_fixture.read_bytes(remove_from_start=2)

    assert response == expected_response
    mock_serial.read_until.assert_called_once()


def test_read_bytes_remove_from_end(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    mock_serial.read_until.return_value = b"test12"
    expected_response = b"test"

    response = device_fixture.read_bytes(remove_from_end=2)

    assert response == expected_response
    mock_serial.read_until.assert_called_once()

    with pytest.raises(ValueError):
        device_fixture.read_bytes(remove_from_end=-1)


def test_read_bytes_remove_from_start_and_end(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    mock_serial.read_until.return_value = b"test\r\n"
    expected_response = b"st"

    response = device_fixture.read_bytes(remove_from_start=2, remove_from_end=2)

    assert response == expected_response
    mock_serial.read_until.assert_called_once_with(expected=b"\r\n", size=None)

    with pytest.raises(IndexError):
        mock_serial.read_until.return_value = b"123456"
        device_fixture.read_bytes(remove_from_start=3, remove_from_end=3)

    with pytest.raises(IndexError):
        mock_serial.read_until.return_value = b"123456789"
        device_fixture.read_bytes(remove_from_start=6, remove_from_end=7)


def test_query(mock_serial, device_fixture):
    """


    Args:
      mock_serial:
      device_fixture:

    Returns:

    """
    for command in ["test", b"test", bytearray("test", "utf-8")]:
        mock_serial.read_until.return_value = b"test"
        response = device_fixture.query(command)

        assert response == "test"
        device_fixture.device.write.assert_called_once_with(b"test")
        device_fixture.device.read_until.assert_called_once_with(expected=b"\r\n", size=None)
        mock_serial.reset_mock()


def test_init_deferred_connection(mock_serial):
    serial_device = SerialDevice(com_port=PORT, connect_hardware=False)

    assert serial_device.device is None
    mock_serial._factory.assert_not_called()


def test_shared_port_reuses_serial_handle(mock_serial):
    device_a = SerialDevice(com_port=PORT, share_port=True)
    device_b = SerialDevice(com_port=PORT, share_port=True)

    assert device_a.device is device_b.device
    assert mock_serial._factory.call_count == 1
    assert SerialDevice._shared_devices


def test_shared_port_different_settings_do_not_reuse_handle(mock_serial):
    device_a = SerialDevice(com_port=PORT, share_port=True, timeout=1.0)
    device_b = SerialDevice(com_port=PORT, share_port=True, timeout=2.0)

    assert device_a.device is not device_b.device
    assert mock_serial._factory.call_count == 2


def test_shared_port_disconnect_releases_registry_entry(mock_serial):
    device_a = SerialDevice(com_port=PORT, share_port=True)
    device_b = SerialDevice(com_port=PORT, share_port=True)

    assert len(SerialDevice._shared_devices) == 1

    device_a.disconnect()
    assert len(SerialDevice._shared_devices) == 1

    device_b.disconnect()
    assert not SerialDevice._shared_devices


def test_shared_port_deferred_connection_reuses_existing_handle(mock_serial):
    eager_device = SerialDevice(com_port=PORT, share_port=True)
    deferred_device = SerialDevice(com_port=PORT, share_port=True, connect_hardware=False)

    assert deferred_device.device is None

    deferred_device.open_device_comm()

    assert deferred_device.device is eager_device.device
    assert mock_serial._factory.call_count == 1



import time
from threading import RLock
from typing import Any, Dict, Optional, Union
from logging import Logger, getLogger

import serial

SERIAL_PARITY: Dict = {
    "none": serial.PARITY_NONE,
    "odd": serial.PARITY_ODD,
    "even": serial.PARITY_EVEN,
    "mark": serial.PARITY_MARK,
    "space": serial.PARITY_SPACE,
}

SERIAL_BYTESIZE: Dict = {
    5: serial.FIVEBITS,
    6: serial.SIXBITS,
    7: serial.SEVENBITS,
    8: serial.EIGHTBITS,
}

SERIAL_STOPBITS: Dict = {
    1: serial.STOPBITS_ONE,
    1.5: serial.STOPBITS_ONE_POINT_FIVE,
    2: serial.STOPBITS_TWO,
}

SERIAL_FACTORY: Dict = {
    "parity": SERIAL_PARITY,
    "bytesize": SERIAL_BYTESIZE,
    "stopbits": SERIAL_STOPBITS,
}


def open_close(func):
    """
    Decorator for managing the opening and closing of the COM port of the device.
    Closes communication first in case it is open (otherwise will raise error com_port is already open)
    Opens communication before sending/receiving commands, then closes in case of communication errors.

    Args:
      func: Function being wrapped

    Returns:
        wrapper: Wrapped function
    """

    def wrapper(self, *args, **kwargs):
        """
        Wrapper function for managing the opening and closing of the COM port of the device.

        Args:
            self: Instance of the SerialDevice (or child) class whose method is being wrapped
            *args: Additional arguments
            **kwargs: Additional keyword arguments

        Returns:
            func_return: Return value(s) of the wrapped function
        """
        self._ensure_device()
        with self._transport_lock():
            self.close_device_comm()  # This is done to avoid opening a com_port that is already open
            self.open_device_comm()
            try:
                func_return = func(self, *args, **kwargs)
            finally:
                self.close_device_comm()
        return func_return

    return wrapper


class SerialDevice:
    category = "SerialDevice"
    ui_fields  = ("com_port", "baudrate", "bytesize", "parity", "stopbits", "timeout", "encoding")
    """ """

    default_eol: bytes = b"\r\n"
    _shared_devices: Dict[Any, Dict[str, Any]] = {}
    _shared_devices_lock = RLock()

    def __init__(
        self,
        com_port: str,
        baudrate: int = 9600,
        bytesize: int = 8,
        parity: str = "none",
        stopbits: int = 1,
        timeout: Optional[float] = None,
        encoding: str = "utf-8",
        connect_hardware: bool = True,
        share_port: bool = False,
        # rs485: bool = False,
        logger: Optional[Logger] = None,
        **kwargs,
    ) -> None:
        """
        SerialDevice is a class for using serial communication with lab instruments. It will be used as a base class for
        other devices.

        If using the RS485 communication protocol, set rs485 to True. Them com_port corresponds to the address.

        Args:
            com_port: comunication port of the device, e.g., "COM1" for Windows, or "/dev/ttyS0" or "/dev/ttySUSB0" for Linux
            baudrate: baudrate of the device
            bytesize: number of data bits
            parity: parity of the device
            stopbits: number of stop bits
            timeout: time to wait for a response from the device
            encoding: encoding of the device
            rs485: if using RS485 communication protocol
            logger: logger to use
            **kwargs: additional keyword arguments for the serial.Serial class

        Returns:
            None
        """
        # ATTN
        # rs485 in SerialDevice is not actually useful since we are not using any advanced settings of RS485 communication
        # if rs485 and (int(com_port) <= 0 or int(com_port) > 255):
        #     raise ValueError("RS485 address must be between 1 and 255.")

        self.com_port: str = com_port
        self.baudrate: int = baudrate
        self.bytesize: int = SERIAL_BYTESIZE[bytesize]
        self.parity: str = SERIAL_PARITY[parity]
        self.stopbits: int = SERIAL_STOPBITS[stopbits]
        self.timeout: float = timeout
        self.encoding: str = encoding
        self.connect_hardware: bool = connect_hardware
        self.share_port: bool = share_port
        self._serial_kwargs: Dict[str, Any] = dict(kwargs)
        self._shared_key: Optional[Any] = None
        self._device_state: Optional[Dict[str, Any]] = None
        self.device = None
        self._instance_lock = RLock()
        # ATTN
        # same as above
        # self.rs485: bool = rs485

        if connect_hardware:
            self._init_device()
        self.logger = logger if logger is not None else getLogger()

        settings = {
            "com_port": self.com_port,
            "baudrate": self.baudrate,
            "parity": self.parity,
            "bytesize": self.bytesize,
            "timeout": self.timeout,
            "stopbits": self.stopbits,
            # "rs485": self.rs485,
            "encoding": self.encoding,
            "connect_hardware": self.connect_hardware,
            "share_port": self.share_port,
            **kwargs,
        }

        self.logger.debug(f"Created SerialDevice with settings: {settings}")

    def _init_device(self, **kwargs) -> None:
        """
        Initialize serial communication with a device.

        Args:
            **kwargs: additional keyword arguments

        Returns:
            None
        """
        if kwargs:
            self._serial_kwargs.update(kwargs)

        if self.share_port:
            self._init_shared_device()
        else:
            self._init_private_device()

        self.close_device_comm()

    @staticmethod
    def _freeze_value(value: Any) -> Any:
        if isinstance(value, dict):
            return tuple(sorted((key, SerialDevice._freeze_value(item)) for key, item in value.items()))
        if isinstance(value, (list, tuple)):
            return tuple(SerialDevice._freeze_value(item) for item in value)
        if isinstance(value, set):
            return tuple(sorted(SerialDevice._freeze_value(item) for item in value))
        return value

    def _shared_device_key(self) -> Any:
        return (
            self.com_port,
            self.baudrate,
            self.parity,
            self.bytesize,
            self.timeout,
            self.stopbits,
            tuple(sorted((key, self._freeze_value(value)) for key, value in self._serial_kwargs.items())),
        )

    def _create_serial_device(self):
        # ATTN: How would Serial class be initialized for RS485?
        return serial.Serial(
            port=self.com_port,
            baudrate=self.baudrate,
            parity=self.parity,
            bytesize=self.bytesize,
            timeout=self.timeout,
            stopbits=self.stopbits,
            **self._serial_kwargs,
        )

    def _init_private_device(self) -> None:
        self.device = self._create_serial_device()

    def _init_shared_device(self) -> None:
        shared_key = self._shared_device_key()
        with self._shared_devices_lock:
            state = self._shared_devices.get(shared_key)
            if state is None:
                state = {
                    "device": self._create_serial_device(),
                    "lock": RLock(),
                    "refcount": 0,
                }
                self._shared_devices[shared_key] = state
            state["refcount"] += 1

        self._shared_key = shared_key
        self._device_state = state
        self.device = state["device"]

    def _ensure_device(self) -> None:
        if self.device is None:
            self._init_device()

    def _transport_lock(self):
        if self.share_port and self._device_state is not None:
            return self._device_state["lock"]
        return self._instance_lock

    def disconnect(self) -> None:
        if self.share_port:
            if self._shared_key is None:
                self.device = None
                self._device_state = None
                return

            with self._shared_devices_lock:
                state = self._shared_devices.get(self._shared_key)
                if state is not None:
                    state["refcount"] -= 1
                    if state["refcount"] <= 0:
                        state["device"].close()
                        self._shared_devices.pop(self._shared_key, None)

            self._shared_key = None
            self._device_state = None
            self.device = None
            return

        if self.device is not None:
            self.device.close()
            self.device = None

    def open_device_comm(self) -> None:
        """
        Closes communication with the device to free the port, then opens communication with the device.
        DO NOT start the device electronically!

        Returns:
            None
        """
        self._ensure_device()
        self.device.close()
        self.device.open()

    def close_device_comm(self) -> None:
        """
        Closes communication with the device.
        DO NOT shut down the device electronically!

        Returns:
            None
        """
        if self.device is None:
            return
        self.device.close()

    @staticmethod
    def _encode(command: Union[str, bytes, bytearray]) -> bytes:
        """
        Encodes the commands to bytes.

        Args:
          command: command to write

        Returns:
            command: command encoded as bytes

        Raises:
            TypeError: If the command type is not str, bytes, or bytearray
        """
        if isinstance(command, str):
            command: bytes = command.encode()
        elif isinstance(command, bytearray):
            command: bytes = bytes(command)
        elif isinstance(command, bytes):
            pass
        else:
            raise TypeError("'command' type must be either  str, bytes, or bytearray.")
        return command

    def write(self, command: Union[str, bytes, bytearray]) -> None:
        """
        Writes the command in bytes to the device

        Args:
          command: command to write

        Returns:
            None
        """
        command_bytes: bytes = self._encode(command)
        self.device.write(command_bytes)

    def read(self, return_bytes: bool = False, **kwargs) -> Union[str, bytes]:
        """
        Reads the response from the device until the end-of-line (EOL) character.

        Args:
          return_bytes: Whether to return the response as bytes or as a string (Default value = False)
          **kwargs: Additional keyword arguments for the read_bytes method

        Returns:
            device response, not including the EOL
        """
        response: bytes = self.read_bytes(**kwargs)

        if not return_bytes:
            response: str = response.decode(self.encoding)

        return response

    def read_bytes(
        self,
        read_until: Optional[Union[str, bytes]] = None,
        num_bytes: int = None,
        remove_from_start: int = 0,
        remove_from_end: Optional[int] = None,
    ) -> bytes:
        # TODO: Clean up docstring
        """
        Reads the response from the device until EOL (b'\r\n')
        num_bytes: max number of bytes
        :return: decoded string, not including the EOL

        Args:
          read_until: Character(s) until which to read the response. If None, then the default EOL is used. (Default value = None)
          num_bytes: Number of bytes to read. If None, then read until read_until. (Default value = None)
          remove_from_start: Number of bytes to remove from the start of the response. (Default value = 0)
          remove_from_end: Number of bytes to remove from the end of the response. (Default value = None)

        Returns:
            Device response in bytes with the specified bytes removed from the start and end

        Raises:
            ValueError: If remove_from_end is a negative integer
            IndexError: If remove_from_start is greater than the length of the full response minus remove_from_end
        """
        if remove_from_end is not None and remove_from_end < 0:
            raise ValueError("remove_from_end must be a positive integer.")

        read_until: Union[str, bytes] = read_until if read_until else self.default_eol
        command_bytes: bytes = self._encode(read_until)
        full_response: bytes = self.device.read_until(expected=command_bytes, size=num_bytes)

        # ATTN: I don't remember why we have this if statement... Why not default remove_from_end to 0?
        if not remove_from_end:
            response = full_response[remove_from_start:]
        else:
            if remove_from_start >= len(full_response) - remove_from_end:
                raise IndexError("End index is less than start index.")

            response = full_response[remove_from_start:-remove_from_end]

        return response

    def query(
        self,
        write_command: Union[str, bytes, bytearray],
        read_delay: float = 0.5,
        return_bytes: bool = False,
        **kwargs,
    ) -> Union[str, bytes]:
        """
        Send command to the device, wait for read_delay seconds, then read the response.

        Args:
          write_command: Command to write
          read_delay: Time to wait before reading the response (Default value = 0.5)
          return_bytes: True to return bytes without decoding to str
          **kwargs: Additional keyword arguments for the read method

        Returns:
            Device response string
        """
        self.write(write_command)
        time.sleep(read_delay)
        return self.read(return_bytes, **kwargs)

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass
